import pathlib
import platform
import sys

import pytest

from ..paths import file_uri_to_path, is_relative, normalized_uri

WIN = platform.system() == "Windows"
HOME = pathlib.Path("~").expanduser()
PY35 = sys.version_info[:2] == (3, 5)


@pytest.mark.skipif(WIN, reason="can't test POSIX paths on Windows")
@pytest.mark.parametrize("root_dir, expected_root_uri", [["~", HOME.as_uri()]])
def test_normalize_posix_path_home(root_dir, expected_root_uri):  # pragma: no cover
    assert normalized_uri(root_dir) == expected_root_uri


@pytest.mark.skipif(WIN, reason="can't test POSIX paths on Windows")
@pytest.mark.parametrize(
    "root, path",
    [["~", "~/a"], ["~", "~/a/../b/"], ["/", "/"], ["/a", "/a/b"], ["/a", "/a/b/../c"]],
)
def test_is_relative_ok(root, path):
    assert is_relative(root, path)


@pytest.mark.skipif(WIN, reason="can't test POSIX paths on Windows")
@pytest.mark.parametrize(
    "root, path",
    [
        ["~", "~/.."],
        ["~", "/"],
        ["/a", "/"],
        ["/a/b", "/a"],
        ["/a/b", "/a/b/.."],
        ["/a", "/a/../b"],
        ["/a", "a//"],
    ],
)
def test_is_relative_not_ok(root, path):
    assert not is_relative(root, path)


@pytest.mark.skipif(not WIN, reason="can't test Windows paths on POSIX")
@pytest.mark.parametrize(
    "root, path",
    [
        ["c:\\Users\\user1", "c:\\Users\\"],
        ["c:\\Users\\user1", "d:\\"],
        ["c:\\Users", "c:\\Users\\.."],
    ],
)
def test_is_relative_not_ok_win(root, path):
    assert not is_relative(root, path)


@pytest.mark.skipif(PY35, reason="can't test non-existent paths on py35")
@pytest.mark.skipif(WIN, reason="can't test POSIX paths on Windows")
@pytest.mark.parametrize(
    "root_dir, expected_root_uri",
    [
        # probably need to try some other things
        [str(HOME / "foo"), (HOME / "foo").as_uri()]
    ],
)
def test_normalize_posix_path_home_subdir(
    root_dir, expected_root_uri
):  # pragma: no cover
    assert normalized_uri(root_dir) == expected_root_uri


@pytest.mark.skipif(not WIN, reason="can't test Windows paths on POSIX")
@pytest.mark.parametrize(
    "root_dir, expected_root_uri",
    [
        ["c:\\Users\\user1", "file:///c:/Users/user1"],
        ["C:\\Users\\user1", "file:///c:/Users/user1"],
        ["//VBOXSVR/shared-folder", "file://vboxsvr/shared-folder/"],
    ],
)
def test_normalize_windows_path_case(root_dir, expected_root_uri):  # pragma: no cover
    try:
        normalized = normalized_uri(root_dir)
    except FileNotFoundError as err:
        if sys.version_info < (3, 12) and root_dir == "//VBOXSVR/shared-folder":
            # see https://github.com/python/cpython/issues/136755
            return
        raise err

    assert normalized == expected_root_uri


@pytest.mark.skipif(WIN, reason="can't test POSIX paths on Windows")
@pytest.mark.parametrize(
    "file_uri, expected_posix_path",
    [
        ["file:///C:/Windows/System32/Drivers/etc", "/C:/Windows/System32/Drivers/etc"],
        ["file:///C:/some%20dir/some%20file.txt", "/C:/some dir/some file.txt"],
        ["file:///home/user/some%20file.txt", "/home/user/some file.txt"],
    ],
)
def test_file_uri_to_path_posix(file_uri, expected_posix_path):  # pragma: no cover
    assert file_uri_to_path(file_uri) == expected_posix_path


@pytest.mark.skipif(not WIN, reason="can't test Windows paths on POSIX")
@pytest.mark.parametrize(
    "file_uri, expected_windows_path",
    [
        # https://github.com/jupyter-lsp/jupyterlab-lsp/pull/305#issuecomment-665996145
        ["file:///C:/Windows/System32/Drivers/etc", "C:/Windows/System32/Drivers/etc"],
        ["file:///C:/some%20dir/some%20file.txt", "C:/some dir/some file.txt"],
    ],
)
def test_file_uri_to_path_windows(file_uri, expected_windows_path):  # pragma: no cover
    assert file_uri_to_path(file_uri) == expected_windows_path



from .config import load_config_schema
from .utils import ShellSpec


class BasedPyrightLanguageServer(ShellSpec):
    key = "basedpyright"
    cmd = "basedpyright-langserver"
    args = ["--stdio"]
    languages = ["python"]
    spec = dict(
        display_name=key,
        mime_types=["text/python", "text/x-ipython"],
        urls=dict(
            home="https://github.com/DetachHead/basedpyright",
            issues="https://github.com/DetachHead/basedpyright/issues",
        ),
        install=dict(
            pip="pip install basedpyright",
            conda="conda install -c conda-forge basedpyright",
        ),
        config_schema=load_config_schema(key),
        requires_documents_on_disk=False,
    )



from .config import load_config_schema
from .utils import ShellSpec

TROUBLESHOOT = """\
Please ensure that texlab executable is in the PATH; \
this should happen automatically when installing texlab from Conda, \
but may require manual configuration of PATH environment variable \
if you compiled texlab from source.

You can ensure check if texlab is in the PATH, by running:

  which texlab

which should return the path to the executable (if found).
"""


class Texlab(ShellSpec):
    cmd = key = "texlab"
    languages = ["tex", "latex"]
    spec = dict(
        display_name="texlab",
        mime_types=["text/x-latex", "text/x-tex"],
        urls=dict(
            home="https://texlab.netlify.app",
            issues="https://github.com/latex-lsp/texlab/issues",
        ),
        install=dict(conda="conda install -c conda-forge texlab chktex"),
        config_schema=load_config_schema(key),
        env=dict(RUST_BACKTRACE="1"),
        troubleshoot=TROUBLESHOOT,
    )



from .config import load_config_schema
from .utils import ShellSpec


class PyreflyLanguageServer(ShellSpec):
    key = cmd = "pyrefly"
    args = ["lsp"]
    languages = ["python"]
    spec = dict(
        display_name="Pyrefly",
        mime_types=["text/python", "text/x-ipython"],
        urls=dict(
            home="https://github.com/facebook/pyrefly",
            issues="https://github.com/facebook/pyrefly/issues",
        ),
        install=dict(
            pip="pip install pyrefly",
            uv="uv add pyrefly",
            conda="conda install -c conda-forge pyrefly",
        ),
        config_schema=load_config_schema(key),
        requires_documents_on_disk=False,
    )



from .config import load_config_schema
from .utils import ShellSpec

TROUBLESHOOT = """\
Please ensure that RScript executable is in the PATH; \
this should happen automatically when using Linux, Mac OS or Conda, \
but will require manual configuration when using the default R installer on Windows.

For more details please consult documentation:
https://cran.r-project.org/bin/windows/base/rw-FAQ.html#Rcmd-is-not-found-in-my-PATH_0021

If Rscript is already in the PATH, you can check whether \
the language server package is properly installed with:

  Rscript -e "cat(system.file(package='languageserver'))"

which should return the path to the installed package.
"""


class RLanguageServer(ShellSpec):
    package = "languageserver"
    key = "r-languageserver"
    cmd = "Rscript"

    @property
    def args(self):
        return ["--slave", "-e", f"{self.package}::run()"]

    @property
    def is_installed_args(self):
        return ["-e", f"cat(system.file(package='{self.package}'))"]

    languages = ["r"]
    spec = dict(
        display_name=key,
        mime_types=["text/x-rsrc"],
        urls=dict(
            home="https://github.com/REditorSupport/languageserver",
            issues="https://github.com/REditorSupport/languageserver/issues",
        ),
        install=dict(
            cran=f'install.packages("{package}")',
            conda="conda install -c conda-forge r-languageserver",
        ),
        config_schema=load_config_schema(key),
        troubleshoot=TROUBLESHOOT,
    )



""" special constants used throughout jupyter_lsp
"""

# the current `entry_point` to use for python-based spec finders
EP_SPEC_V1 = "jupyter_lsp_spec_v1"

# the current `entry_point`s to use for python-based listeners
EP_LISTENER_ALL_V1 = "jupyter_lsp_listener_all_v1"
EP_LISTENER_CLIENT_V1 = "jupyter_lsp_listener_client_v1"
EP_LISTENER_SERVER_V1 = "jupyter_lsp_listener_server_v1"

# jupyter*config.d where language_servers can be defined
APP_CONFIG_D_SECTIONS = ["_", "_notebook_", "_server_"]



import os

import pytest


def test_serverextension_path(app):
    import jupyter_lsp

    paths = jupyter_lsp._jupyter_server_extension_paths()
    for path in paths:
        assert __import__(path["module"])


def test_serverextension(app):
    app.initialize(
        ["--ServerApp.jpserver_extensions={'jupyter_lsp.serverextension': True}"]
    )
    assert app.language_server_manager
    found_lsp = False
    for r in app.web_app.default_router.rules:
        for rr in r.target.rules:
            if "/lsp/" in str(rr.matcher.regex):
                found_lsp = True

    assert found_lsp, "apparently didn't install the /lsp/ route"


def test_default_virtual_documents_dir(app):
    app.initialize(
        ["--ServerApp.jpserver_extensions={'jupyter_lsp.serverextension': True}"]
    )
    assert app.language_server_manager.virtual_documents_dir == ".virtual_documents"


def test_virtual_documents_dir_config(app):
    custom_dir = ".custom_virtual_dir"
    app.initialize(
        [
            "--ServerApp.jpserver_extensions={'jupyter_lsp.serverextension': True}",
            "--ServerApp.LanguageServerManager.virtual_documents_dir=" + custom_dir,
        ]
    )
    assert app.language_server_manager.virtual_documents_dir == custom_dir


def test_virtual_documents_dir_env(app):
    os.environ["JP_LSP_VIRTUAL_DIR"] = custom_dir = ".custom_virtual_dir"
    app.initialize(
        ["--ServerApp.jpserver_extensions={'jupyter_lsp.serverextension': True}"]
    )
    assert app.language_server_manager.virtual_documents_dir == custom_dir
    page_config = app.web_app.settings["page_config_data"]
    assert page_config["virtualDocumentsUri"].endswith(custom_dir)


@pytest.mark.parametrize("invalid_path", ["", "."])
def test_virtual_documents_dir_env_empty(app, invalid_path):
    os.environ["JP_LSP_VIRTUAL_DIR"] = invalid_path
    app.initialize(
        ["--ServerApp.jpserver_extensions={'jupyter_lsp.serverextension': True}"]
    )
    page_config = app.web_app.settings["page_config_data"]
    assert page_config["virtualDocumentsUri"].endswith("/.virtual_documents")



import json
import os
import pathlib
import shutil
from pathlib import Path
from typing import Text

from jupyter_server.serverapp import ServerApp
from pytest import fixture
from tornado.httpserver import HTTPRequest
from tornado.httputil import HTTPServerRequest
from tornado.queues import Queue
from tornado.web import Application

# local imports
from jupyter_lsp import LanguageServerManager
from jupyter_lsp.constants import APP_CONFIG_D_SECTIONS
from jupyter_lsp.handlers import LanguageServersHandler, LanguageServerWebSocketHandler

# these should always be available in a test environment
KNOWN_SERVERS = [
    "bash-language-server",
    "dockerfile-language-server-nodejs",
    "typescript-language-server",
    "pylsp",
    "unified-language-server",
    "sql-language-server",
    "vscode-css-languageserver-bin",
    "vscode-html-languageserver-bin",
    "vscode-json-languageserver-bin",
    "yaml-language-server",
]

CMD_BASED_SERVERS = {
    "Rscript": ["r-languageserver"],
    "texlab": ["texlab"],
    "jedi-language-server": ["jedi-language-server"],
    "julia": ["julia-language-server"],
}

KNOWN_SERVERS += sum(
    [langs for cmd, langs in CMD_BASED_SERVERS.items() if shutil.which(cmd)], []
)

KNOWN_UNKNOWN_SERVERS = ["foo-language-server"]


def extra_node_roots():
    root = Path(os.environ.get("JLSP_TEST_ROOT") or Path.cwd())
    return dict(extra_node_roots=[str(root)] if root else [])


@fixture
def manager() -> LanguageServerManager:
    return LanguageServerManager(**extra_node_roots())


@fixture
def echo_spec():
    return {"argv": ["echo", "no server here"], "languages": ["klingon"], "version": 2}


@fixture
def echo_conf_json(echo_spec) -> str:
    return json.dumps(
        {"LanguageServerManager": {"language_servers": {"_echo_": echo_spec}}},
        indent=2,
        sort_keys=True,
    )


@fixture(params=sorted(APP_CONFIG_D_SECTIONS))
def app_config_d(request, tmp_path, monkeypatch) -> pathlib.Path:
    conf_d = tmp_path / f"jupyter{request.param}config.d"
    conf_d.mkdir()
    monkeypatch.setenv("JUPYTER_CONFIG_PATH", f"{tmp_path}")
    return conf_d


@fixture(params=sorted(KNOWN_SERVERS))
def known_server(request):
    return request.param


@fixture(params=sorted(KNOWN_UNKNOWN_SERVERS))
def known_unknown_server(request):
    return request.param


@fixture
def handlers(manager):
    ws_handler = MockWebsocketHandler()
    ws_handler.initialize(manager)
    handler = MockHandler()
    handler.initialize(manager)
    return handler, ws_handler


@fixture
def jsonrpc_init_msg():
    return json.dumps(
        {
            "id": 0,
            "jsonrpc": "2.0",
            "method": "initialize",
            "params": {
                "capabilities": {
                    # see: https://github.com/julia-vscode/LanguageServer.jl/issues/1008
                    # LanguageServer.jl assumes that it is not missing
                    "workspace": {"didChangeConfiguration": {}},
                    # LanguageServer.jl assumes that it is not missing
                    "textDocument": {},
                },
                "initializationOptions": None,
                "processId": None,
                "rootUri": pathlib.Path(__file__).parent.as_uri(),
                "workspaceFolders": None,
            },
        }
    )


@fixture
def app():
    return MockServerApp()


# mocks
class MockWebsocketHandler(LanguageServerWebSocketHandler):
    _messages_wrote = None  # type: Queue
    _ping_sent = None  # type: bool

    def __init__(self):
        self.request = HTTPServerRequest()
        self.application = Application()

    def initialize(self, manager):
        super().initialize(manager)
        self._messages_wrote = Queue()
        self._ping_sent = False

    def write_message(self, message: Text) -> None:  # type: ignore
        self.log.warning("write_message %s", message)
        self._messages_wrote.put_nowait(message)

    def send_ping(self):
        self._ping_sent = True


class MockHandler(LanguageServersHandler):
    _payload = None
    _jupyter_current_user = "foo"  # type:ignore[assignment]

    def __init__(self):
        self.request = HTTPRequest("GET")
        self.application = Application()

    def finish(self, payload):
        self._payload = payload


class MockServerApp(ServerApp):
    pass



import os
import shutil
import sys
from pathlib import Path
from subprocess import check_output
from typing import List, Text, Union

from ..schema import SPEC_VERSION
from ..types import (
    KeyedLanguageServerSpecs,
    LanguageServerManagerAPI,
    LanguageServerSpec,
    SpecBase,
    Token,
)

# helper scripts for known tricky language servers
HELPERS = Path(__file__).parent / "helpers"

# when building docs, let all specs go through
BUILDING_DOCS = os.environ.get("JUPYTER_LSP_BUILDING_DOCS") is not None


class ShellSpec(SpecBase):  # pragma: no cover
    """Helper for a language server spec for executables on $PATH in the
    notebook server environment.
    """

    cmd = ""

    # [optional] arguments passed to `cmd` which upon execution should print
    # out a non-empty string if the the required language server package
    # is installed, or nothing if it is missing and user action is required.
    is_installed_args: List[Token] = []

    def is_installed(self, mgr: LanguageServerManagerAPI) -> bool:
        cmd = self.solve()

        if not cmd:
            return False

        if not self.is_installed_args:
            return bool(cmd)
        else:
            check_result = check_output([cmd, *self.is_installed_args]).decode(
                encoding="utf-8"
            )
            return check_result != ""

    def solve(self) -> Union[str, None]:
        for ext in ["", ".cmd", ".bat", ".exe"]:
            cmd = shutil.which(self.cmd + ext)
            if cmd:
                break
        return cmd

    def __call__(self, mgr: LanguageServerManagerAPI) -> KeyedLanguageServerSpecs:
        cmd = self.solve()

        spec = dict(self.spec)

        if not cmd:
            troubleshooting = [f"{self.cmd} not found."]
            if "troubleshoot" in spec:
                troubleshooting.append(spec["troubleshoot"])
            spec["troubleshoot"] = "\n\n".join(troubleshooting)

        if not cmd and BUILDING_DOCS:  # pragma: no cover
            cmd = self.cmd

        return {
            self.key: {
                "argv": [cmd, *self.args] if cmd else [self.cmd, *self.args],
                "languages": self.languages,
                "version": SPEC_VERSION,
                **spec,
            }
        }


class PythonModuleSpec(SpecBase):
    """Helper for a python-based language server spec in the notebook server
    environment
    """

    python_module = ""

    def is_installed(self, mgr: LanguageServerManagerAPI) -> bool:
        spec = self.solve()

        if not spec:
            return False

        if not spec.origin:  # pragma: no cover
            return False

        return True

    def solve(self):
        return __import__("importlib").util.find_spec(self.python_module)

    def __call__(self, mgr: LanguageServerManagerAPI) -> KeyedLanguageServerSpecs:
        is_installed = self.is_installed(mgr)

        return {
            self.key: {
                "argv": (
                    [sys.executable, "-m", self.python_module, *self.args]
                    if is_installed
                    else []
                ),
                "languages": self.languages,
                "version": SPEC_VERSION,
                **self.spec,
            }
        }


class NodeModuleSpec(SpecBase):
    """Helper for a nodejs-based language server spec in one of several
    node_modules
    """

    node_module = ""
    script: List[Text] = []

    def is_installed(self, mgr: LanguageServerManagerAPI) -> bool:
        node_module = self.solve(mgr)
        return bool(node_module)

    def solve(self, mgr: LanguageServerManagerAPI):
        return mgr.find_node_module(self.node_module, *self.script)

    def __call__(self, mgr: LanguageServerManagerAPI) -> KeyedLanguageServerSpecs:
        node_module = self.solve(mgr)

        spec = dict(self.spec)

        troubleshooting = ["Node.js is required to install this server."]
        if "troubleshoot" in spec:  # pragma: no cover
            troubleshooting.append(spec["troubleshoot"])
        spec["troubleshoot"] = "\n\n".join(troubleshooting)

        is_installed = self.is_installed(mgr)

        return {
            self.key: {
                "argv": ([mgr.nodejs, node_module, *self.args] if is_installed else []),
                "languages": self.languages,
                "version": SPEC_VERSION,
                **spec,
            }
        }


# these are not desirable to publish to the frontend
# and will be replaced with the simplest schema-compliant values
SKIP_JSON_SPEC = {"argv": [""], "debug_argv": [""], "env": {}}


def censored_spec(spec: LanguageServerSpec) -> LanguageServerSpec:
    return {k: SKIP_JSON_SPEC.get(k, v) for k, v in spec.items()}



import asyncio
import re

import pytest
import traitlets
from tornado.queues import Queue

from jupyter_lsp import lsp_message_listener


@pytest.mark.parametrize("bad_string", ["not-a-function", "jupyter_lsp.__version__"])
@pytest.mark.asyncio
async def test_listener_bad_traitlets(bad_string, handlers):
    handler, ws_handler = handlers
    manager = handler.manager

    with pytest.raises(traitlets.TraitError):
        manager.all_listeners = [bad_string]


@pytest.mark.asyncio
async def test_listeners(known_server, handlers, jsonrpc_init_msg):
    """will some listeners listen?"""
    handler, ws_handler = handlers
    manager = handler.manager

    manager.all_listeners = ["jupyter_lsp.tests.listener.dummy_listener"]

    manager.initialize()
    manager._listeners["client"] = []  # hide predefined client listeners

    assert len(manager._listeners["all"]) == 1

    dummy_listener = manager._listeners["all"][0]
    assert re.match(
        (
            "<MessageListener listener=<function dummy_listener at .*?>,"
            " method=None, language_server=None>"
        ),
        repr(dummy_listener),
    )

    handler_listened = Queue()
    server_listened = Queue()
    all_listened = Queue()

    # some client listeners
    @lsp_message_listener("client", language_server=known_server, method="initialize")
    async def client_listener(scope, message, language_server, manager):
        await handler_listened.put(message)

    @lsp_message_listener("client", method=r"not-a-method")
    async def other_client_listener(
        scope, message, language_server, manager
    ):  # pragma: no cover
        await handler_listened.put(message)
        raise NotImplementedError("shouldn't get here")

    # some server listeners
    @lsp_message_listener("server", language_server=None, method=None)
    async def server_listener(scope, message, language_server, manager):
        await server_listened.put(message)

    @lsp_message_listener("server", language_server=r"not-a-language-server")
    async def other_server_listener(
        scope, message, language_server, manager
    ):  # pragma: no cover
        await handler_listened.put(message)
        raise NotImplementedError("shouldn't get here")

    # an all listener
    @lsp_message_listener("all")
    async def all_listener(
        scope, message, language_server, manager
    ):  # pragma: no cover
        await all_listened.put(message)

    assert len(manager._listeners["server"]) == 2
    assert len(manager._listeners["client"]) == 2
    assert len(manager._listeners["all"]) == 2

    await ws_handler.open(known_server)

    await ws_handler.on_message(jsonrpc_init_msg)

    results = await asyncio.wait_for(
        asyncio.gather(
            handler_listened.get(),
            server_listened.get(),
            all_listened.get(),
            all_listened.get(),
            return_exceptions=True,
        ),
        240 if known_server == "julia-language-server" else 20,
    )
    assert all([isinstance(res, dict) for res in results])

    ws_handler.on_close()

    handler_listened.task_done()
    server_listened.task_done()
    all_listened.task_done()
    all_listened.task_done()

    [
        manager.unregister_message_listener(listener)
        for listener in [
            client_listener,
            other_client_listener,
            server_listener,
            other_server_listener,
            all_listener,
        ]
    ]

    assert not manager._listeners["server"]
    assert not manager._listeners["client"]
    assert len(manager._listeners["all"]) == 1



""" A configurable frontend for stdio-based Language Servers
"""

import asyncio
import os
import sys
import traceback
from typing import Dict, Text, Tuple, cast

# See compatibility note on `group` keyword in
# https://docs.python.org/3/library/importlib.metadata.html#entry-points
if sys.version_info < (3, 10):  # pragma: no cover
    from importlib_metadata import entry_points
else:  # pragma: no cover
    from importlib.metadata import entry_points

from jupyter_core.paths import jupyter_config_path
from jupyter_server.services.config import ConfigManager

try:
    from jupyter_server.transutils import _i18n as _
except ImportError:  # pragma: no cover
    from jupyter_server.transutils import _

from traitlets import Bool
from traitlets import Dict as Dict_
from traitlets import Instance
from traitlets import List as List_
from traitlets import Unicode, default

from .constants import (
    APP_CONFIG_D_SECTIONS,
    EP_LISTENER_ALL_V1,
    EP_LISTENER_CLIENT_V1,
    EP_LISTENER_SERVER_V1,
    EP_SPEC_V1,
)
from .schema import LANGUAGE_SERVER_SPEC_MAP
from .session import LanguageServerSession
from .trait_types import LoadableCallable, Schema
from .types import (
    KeyedLanguageServerSpecs,
    LanguageServerManagerAPI,
    MessageScope,
    SpecBase,
    SpecMaker,
)


class LanguageServerManager(LanguageServerManagerAPI):
    """Manage language servers"""

    conf_d_language_servers = Schema(  # type:ignore[assignment]
        validator=LANGUAGE_SERVER_SPEC_MAP,
        help=_("extra language server specs, keyed by implementation, from conf.d"),
    )  # type: KeyedLanguageServerSpecs

    language_servers = Schema(  # type:ignore[assignment]
        validator=LANGUAGE_SERVER_SPEC_MAP,
        help=_("a dict of language server specs, keyed by implementation"),
    ).tag(
        config=True
    )  # type: KeyedLanguageServerSpecs

    autodetect: bool = Bool(  # type:ignore[assignment]
        True, help=_("try to find known language servers in sys.prefix (and elsewhere)")
    ).tag(config=True)

    sessions: Dict[Tuple[Text], LanguageServerSession] = (
        Dict_(  # type:ignore[assignment]
            trait=Instance(LanguageServerSession),
            default_value={},
            help="sessions keyed by language server name",
        )
    )

    virtual_documents_dir = Unicode(
        help="""Path to virtual documents relative to the content manager root
        directory.

        Its default value can be set with JP_LSP_VIRTUAL_DIR and fallback to
        '.virtual_documents'.
        """
    ).tag(config=True)

    _ready = Bool(
        help="""Whether the manager has been initialized""", default_value=False
    )

    all_listeners = List_(  # type:ignore[var-annotated]
        trait=LoadableCallable  # type:ignore[arg-type]
    ).tag(config=True)
    server_listeners = List_(  # type:ignore[var-annotated]
        trait=LoadableCallable  # type:ignore[arg-type]
    ).tag(config=True)
    client_listeners = List_(  # type:ignore[var-annotated]
        trait=LoadableCallable  # type:ignore[arg-type]
    ).tag(config=True)

    @default("language_servers")
    def _default_language_servers(self):
        return {}

    @default("virtual_documents_dir")
    def _default_virtual_documents_dir(self):
        return os.getenv("JP_LSP_VIRTUAL_DIR", None) or ".virtual_documents"

    @default("conf_d_language_servers")
    def _default_conf_d_language_servers(self) -> KeyedLanguageServerSpecs:
        language_servers: KeyedLanguageServerSpecs = {}

        manager = ConfigManager(read_config_path=jupyter_config_path())

        for app in APP_CONFIG_D_SECTIONS:
            language_servers.update(
                **manager.get(f"jupyter{app}config")
                .get(self.__class__.__name__, {})
                .get("language_servers", {})
            )

        return language_servers

    def __init__(self, **kwargs: Dict):
        """Before starting, perform all necessary configuration"""
        self.all_language_servers: KeyedLanguageServerSpecs = {}
        self._language_servers_from_config: KeyedLanguageServerSpecs = {}
        super().__init__(**kwargs)

    def initialize(self, *args, **kwargs):
        self.init_language_servers()
        self.init_listeners()
        self.init_sessions()
        self._ready = True

    async def ready(self):
        while not self._ready:  # pragma: no cover
            await asyncio.sleep(0.1)
        return True

    def init_language_servers(self) -> None:
        """determine the final language server configuration."""
        # copy the language servers before anybody monkeys with them
        self._language_servers_from_config = dict(self.language_servers)
        self.language_servers = self._collect_language_servers(only_installed=True)
        self.all_language_servers = self._collect_language_servers(only_installed=False)

    def _collect_language_servers(
        self, only_installed: bool
    ) -> KeyedLanguageServerSpecs:
        language_servers: KeyedLanguageServerSpecs = {}

        language_servers_from_config = dict(self._language_servers_from_config)
        language_servers_from_config.update(self.conf_d_language_servers)

        if self.autodetect:
            language_servers.update(
                self._autodetect_language_servers(only_installed=only_installed)
            )

        # restore config
        language_servers.update(language_servers_from_config)

        # coalesce the servers, allowing a user to opt-out by specifying `[]`
        return {key: spec for key, spec in language_servers.items() if spec.get("argv")}

    def init_sessions(self):
        """create, but do not initialize all sessions"""
        sessions = {}
        for language_server, spec in self.language_servers.items():
            sessions[language_server] = LanguageServerSession(
                language_server=language_server, spec=spec, parent=self
            )
        self.sessions = sessions

    def init_listeners(self):
        """register traitlets-configured listeners"""

        scopes = {
            MessageScope.ALL: [self.all_listeners, EP_LISTENER_ALL_V1],
            MessageScope.CLIENT: [self.client_listeners, EP_LISTENER_CLIENT_V1],
            MessageScope.SERVER: [self.server_listeners, EP_LISTENER_SERVER_V1],
        }
        for scope, trt_ep in scopes.items():
            listeners, entry_point = trt_ep

            for ept in entry_points(group=entry_point):  # pragma: no cover
                try:
                    listeners.append(ept.load())
                except Exception as err:
                    self.log.warning("Failed to load entry point %s: %s", ept.name, err)

            for listener in listeners:
                self.__class__.register_message_listener(scope=scope.value)(listener)

    def subscribe(self, handler):
        """subscribe a handler to session, or sta"""
        session = self.sessions.get(handler.language_server)

        if session is None:
            self.log.error(
                "[{}] no session: handler subscription failed".format(
                    handler.language_server
                )
            )
            return

        session.handlers = set([handler]) | session.handlers

    async def on_client_message(self, message, handler):
        await self.wait_for_listeners(
            MessageScope.CLIENT, message, handler.language_server
        )
        session = self.sessions.get(handler.language_server)

        if session is None:
            self.log.error(
                "[{}] no session: client message dropped".format(
                    handler.language_server
                )
            )
            return

        session.write(message)

    async def on_server_message(self, message, session):
        language_servers = [
            ls_key for ls_key, sess in self.sessions.items() if sess == session
        ]

        for language_servers in language_servers:
            await self.wait_for_listeners(
                MessageScope.SERVER, message, language_servers
            )

        for handler in session.handlers:
            handler.write_message(message)

    def unsubscribe(self, handler):
        session = self.sessions.get(handler.language_server)

        if session is None:
            self.log.error(
                "[{}] no session: handler unsubscription failed".format(
                    handler.language_server
                )
            )
            return

        session.handlers = [h for h in session.handlers if h != handler]

    def _autodetect_language_servers(self, only_installed: bool):
        _entry_points = None

        try:
            _entry_points = entry_points(group=EP_SPEC_V1)
        except Exception:  # pragma: no cover
            self.log.exception("Failed to load entry_points")

        skipped_servers = []

        for ep in _entry_points or []:
            try:
                spec_finder: SpecMaker = ep.load()
            except Exception as err:  # pragma: no cover
                self.log.warning(
                    _("Failed to load language server spec finder `{}`: \n{}").format(
                        ep.name, err
                    )
                )
                continue

            try:
                if only_installed:
                    if hasattr(spec_finder, "is_installed"):
                        spec_finder_from_base = cast(SpecBase, spec_finder)
                        if not spec_finder_from_base.is_installed(self):
                            skipped_servers.append(ep.name)
                            continue
                specs = spec_finder(self) or {}
            except Exception as err:  # pragma: no cover
                self.log.warning(
                    _(
                        "Failed to fetch commands from language server spec finder"
                        " `{}`:\n{}"
                    ).format(ep.name, err)
                )
                traceback.print_exc()

                continue

            errors = list(LANGUAGE_SERVER_SPEC_MAP.iter_errors(specs))

            if errors:  # pragma: no cover
                self.log.warning(
                    _(
                        "Failed to validate commands from language server spec finder"
                        " `{}`:\n{}"
                    ).format(ep.name, errors)
                )
                continue

            for key, spec in specs.items():
                yield key, spec

        if skipped_servers:
            self.log.info(
                _("Skipped non-installed server(s): {}").format(
                    ", ".join(skipped_servers)
                )
            )


# the listener decorator
lsp_message_listener = LanguageServerManager.register_message_listener  # noqa



from .utils import NodeModuleSpec


class VSCodeJSONLanguageServer(NodeModuleSpec):
    node_module = key = "vscode-json-languageserver-bin"
    script = ["jsonServerMain.js"]
    args = ["--stdio"]
    languages = ["json"]
    spec = dict(
        display_name=key,
        mime_types=["application/json", "application/x-json", "application/ld+json"],
        urls=dict(
            home="https://github.com/vscode-langservers/{}".format(key),
            issues="https://github.com/vscode-langservers/{}/issues".format(key),
        ),
        install=dict(
            npm="npm install --save-dev {}".format(key),
            yarn="yarn add --dev {}".format(key),
            jlpm="jlpm add --dev {}".format(key),
        ),
    )
